@@ -19,14 +19,16 @@ from typing import (
     Optional,
     Union,
 )
+import warnings
 
+from ... import _repr
 from ...prim import Party
 from .exc import ConfigError
 
 if sys.version_info >= (3, 8):
-    from typing import Protocol, runtime_checkable
+    from typing import Literal, Protocol, runtime_checkable
 else:
-    from typing_extensions import Protocol, runtime_checkable
+    from typing_extensions import Literal, Protocol, runtime_checkable
 
 __all__ = [
     "AccessConfig",
@@ -170,12 +172,19 @@ class AccessConfig(Protocol):
         """
         raise NotImplementedError
 
+    @ledger_id.setter
+    def ledger_id(self, value: Optional[str]) -> None:
+        # note; this is settable purely to support Daml V1-based workflows where ledger ID needss
+        # to be remembered locally because most interactions with Daml V1 ledgers require
+        # ledger ID to be supplied
+        raise NotImplementedError
+
     @property
-    def application_name(self) -> str:
+    def application_name(self) -> Optional[str]:
         """
         The application name.
 
-        :type: str
+        :type: Optional[str]
         """
         raise NotImplementedError
 
@@ -188,6 +197,21 @@ class AccessConfig(Protocol):
         """
         raise NotImplementedError
 
+    @property
+    def token_version(self) -> "Optional[Literal[1, 2]]":
+        """
+        The version of the token supplied at configuration time, as provided by a signing authority
+        that is trusted by the server.
+
+        This parameter is used at connection initialization in order to bootstrap local state.
+
+        If access parameters are supplied instead of a token, then the version of the token is
+        ``None``.
+
+        :type: int | None
+        """
+        raise NotImplementedError
+
 
 class TokenBasedAccessConfig(AccessConfig):
     """
@@ -195,13 +219,18 @@ class TokenBasedAccessConfig(AccessConfig):
     party rights, the application name, and ledger ID are all derived off of the token.
     """
 
-    def __init__(self, oauth_token: str):
+    _token_version: Literal[1, 2]
+
+    def __init__(self, oauth_token: "Union[bytes, str]"):
         """
         Initialize a token-based access configuration.
 
         :param oauth_token: The initial value of the bearer token.
         """
-        self.token = oauth_token
+        if isinstance(oauth_token, bytes):
+            self.token = oauth_token.decode("ascii")
+        else:
+            self.token = oauth_token
 
     @property
     def token(self) -> str:
@@ -214,17 +243,45 @@ class TokenBasedAccessConfig(AccessConfig):
     @token.setter
     def token(self, value: str) -> None:
         self._token = value
-        claims = decode_token(self._token)
+        claims = decode_token_claims(self._token)
 
-        read_as = frozenset(claims.get("readAs", ()))
-        act_as = frozenset(claims.get("actAs", ()))
+        v1_claims = claims.get(DamlLedgerApiNamespace)
+        if "daml_ledger_api" in claims.get("scope", "").split(" "):
+            # "scope": "daml_ledegr_api" is present, which makes it a Daml V2 token
+            self._ledger_id = None
+            self._application_name = None
+            self._token_version = 2
+
+        elif v1_claims is not None:
+            # we found Daml V1 claims, so assume it's a Daml V1 token
+            self._set(
+                read_as=frozenset(v1_claims.get("readAs", ())),
+                act_as=frozenset(v1_claims.get("actAs", ())),
+                admin=bool(v1_claims.get("admin", False)),
+            )
+            self._ledger_id = v1_claims.get("ledgerId", None)
+            self._application_name = v1_claims.get("applicationId", None)
+            self._token_version = 1
+
+        else:
+            # we're not _entirely_ sure what kind of token it is; assume it's 2
+            self._ledger_id = None
+            self._application_name = None
+            self._token_version = 2
+
+    def _set(self, *, read_as: Collection[Party], act_as: Collection[Party], admin: bool):
+        """
+        Set the values of this :class:`TokenBasedAccessConfig`.
+
+        This is not a public API, and subject to change at any time.
+        """
+        read_as = frozenset(read_as)
+        act_as = frozenset(act_as)
 
         self._act_as = act_as
         self._read_only_as = read_as - act_as
         self._read_as = read_as.union(act_as)
-        self._admin = bool(claims.get("admin", False))
-        self._ledger_id = claims.get("ledgerId", None)
-        self._application_name = claims.get("applicationId", None)
+        self._admin = admin
 
     @property
     def read_as(self) -> AbstractSet[Party]:
@@ -246,10 +303,40 @@ class TokenBasedAccessConfig(AccessConfig):
     def ledger_id(self) -> Optional[str]:
         return self._ledger_id
 
+    @ledger_id.setter
+    def ledger_id(self, value: Optional[str]) -> None:
+        """
+        Set ledger ID.
+
+        Unlike the other properties derived from the token, this is settable here to support
+        Daml V1-based workflows where the ledger ID isn't necessarily in the token, but still
+        needs to be known locally in order to supply it to gRPC commands.
+        """
+        self._ledger_id = value
+
     @property
-    def application_name(self) -> str:
+    def application_name(self) -> Optional[str]:
         return self._application_name
 
+    @property
+    def token_version(self) -> "Literal[1, 2]":
+        return self._token_version
+
+    def __repr__(self):
+        s = [
+            f"read_as={_repr.list(self.read_as)}",
+            f"act_as={_repr.list(self.act_as)}",
+        ]
+        if self.admin:
+            s.append("admin=True")
+        if self.ledger_id is not None:
+            s.append(f"ledger_id={_repr.str(self.ledger_id)}")
+        if self.application_name is not None:
+            s.append(f"application_name={_repr.str(self.application_name)}")
+        s.append(f"token_version={self.token_version}")
+
+        return f"{self.__class__.__name__}({', '.join(s)})"
+
 
 class TokenFileBasedAccessConfig(TokenBasedAccessConfig):
     def __init__(self, oauth_token_file: str):
@@ -299,13 +386,13 @@ class PropertyBasedAccessConfig(AccessConfig):
         self._application_name = application_name or "dazl-client"
 
     @property
-    def token(self):
+    def token(self) -> str:
         """
         Produces a token without signing, utilizing our parameters.
         """
         return encode_unsigned_token(
             self.read_as, self.act_as, self.ledger_id, self.application_name, self.admin
-        )
+        ).decode("ascii")
 
     @property
     def ledger_id(self) -> "Optional[str]":
@@ -371,6 +458,14 @@ class PropertyBasedAccessConfig(AccessConfig):
     def admin(self, value: bool) -> None:
         self._admin = value
 
+    @property
+    def token_version(self) -> None:
+        """
+        Return ``None``, because any token that we are supplying is purely for identification
+        purposes and not really a valid server-side issued token.
+        """
+        return None
+
 
 def parties(p: Union[None, Party, Collection[Party]]) -> Collection[Party]:
     if p is None:
@@ -385,39 +480,51 @@ DamlLedgerApiNamespace = "https://daml.com/ledger-api"
 
 
 def decode_token(token: str) -> Mapping[str, Any]:
-    components = token.split(".", 3)
-    if len(components) != 3:
-        raise ValueError("not a JWT")
-
-    pad_bytes = "=" * (-len(components[1]) % 4)
-    claim_str = base64.urlsafe_b64decode(components[1] + pad_bytes)
-    claims = json.loads(claim_str)
+    warnings.warn("decode_token is deprecated; use decode_token_claims instead", DeprecationWarning)
+    claims = decode_token_claims(token)
     claims_dict = claims.get(DamlLedgerApiNamespace)
     if claims_dict is None:
         raise ValueError(f"JWT is missing claim namespace: {DamlLedgerApiNamespace!r}")
     return claims_dict
 
 
+def decode_token_claims(token: str) -> "Mapping[str, Any]":
+    """
+    Decode the claims section from a JSON Web Token (JWT).
+
+    Note that the signature is NOT verified; this is the responsibility of the caller!
+    """
+    components = token.split(".", 3)
+    if len(components) != 3:
+        raise ValueError("not a JWT")
+
+    pad_bytes = "=" * (-len(components[1]) % 4)
+    claim_str = base64.urlsafe_b64decode(components[1] + pad_bytes)
+    return json.loads(claim_str)
+
+
 def encode_unsigned_token(
-    read_as: Collection[Party],
-    act_as: Collection[Party],
-    ledger_id: str,
-    application_id: str,
+    read_as: "Optional[Collection[Party]]",
+    act_as: "Optional[Collection[Party]]",
+    ledger_id: "Optional[str]",
+    application_id: "Optional[str]",
     admin: bool = True,
 ) -> bytes:
     header = {
         "alg": "none",
         "typ": "JWT",
     }
-    payload = {
-        DamlLedgerApiNamespace: {
-            "ledgerId": ledger_id,
-            "applicationId": application_id,
-            "actAs": sorted(act_as),
-            "readAs": sorted(read_as),
-            "admin": admin,
-        }
-    }
+    payload = {DamlLedgerApiNamespace: {}}  # type: Mapping[str, Any]
+    if ledger_id is not None:
+        payload[DamlLedgerApiNamespace]["ledgerId"] = ledger_id
+    if application_id is not None:
+        payload[DamlLedgerApiNamespace]["applicationId"] = application_id
+    if act_as is not None:
+        payload[DamlLedgerApiNamespace]["actAs"] = sorted(act_as)
+    if read_as is not None:
+        payload[DamlLedgerApiNamespace]["readAs"] = sorted(read_as)
+    if admin:
+        payload[DamlLedgerApiNamespace]["admin"] = admin
 
     return (
         base64.urlsafe_b64encode(json.dumps(header).encode("utf-8")).rstrip(b"=")
