@@ -12,10 +12,12 @@ from typing import (
     AsyncIterable,
     Collection,
     Dict,
+    List,
     Mapping,
     Optional,
     Sequence,
     Union,
+    cast,
 )
 import warnings
 
@@ -32,6 +34,8 @@ from ...prim import ContractData, ContractId, Party, datetime_to_timestamp, to_p
 from ...query import Filter, Queries, Query, parse_query
 from .._offsets import END, End, LedgerOffsetRange, from_offset_until_forever
 from ..api_types import (
+    ActAs,
+    Admin,
     ArchiveEvent,
     Boundary,
     Command,
@@ -40,10 +44,13 @@ from ..api_types import (
     ExerciseResponse,
     ParticipantMeteringReport,
     PartyInfo,
+    ReadAs,
+    Right,
     User,
+    Version,
 )
 from ..config import Config
-from ..config.access import PropertyBasedAccessConfig
+from ..config.access import PropertyBasedAccessConfig, TokenBasedAccessConfig
 from ..errors import ProtocolWarning, _allow_cancel, _translate_exceptions
 from .channel import create_channel
 from .codec_aio import Codec
@@ -93,16 +100,33 @@ class Connection(aio.Connection):
         Does final validation of the token, including possibly fetching the ledger ID if it is not
         yet known.
         """
-        if not self._config.access.ledger_id:
-            # most calls require a ledger ID; if it wasn't supplied as part of our token or we were
-            # never given a token in the first place, fetch the ledger ID from the destination
+        if self._config.access.token_version == 2:
+            # Daml 2.0 tokens do not contain party information, so an extra call to the server is
+            # required in order to resolve our current set of rights
+            if hasattr(self._config.access, "_set"):
+                admin = False
+                read_as = []  # type: List[Party]
+                act_as = []  # type: List[Party]
+                for right in await self.list_user_rights():
+                    if right == Admin:
+                        admin = True
+                    elif isinstance(right, (ReadAs, ActAs)):
+                        read_as.append(right.party)
+
+                # noinspection PyProtectedMember
+                cast(TokenBasedAccessConfig, self._config.access)._set(
+                    read_as=read_as, act_as=act_as, admin=admin
+                )
+
+        elif not self._config.access.ledger_id:
+            # most calls to Daml 1.x ledgers require a ledger ID; if it wasn't supplied as part of
+            # our token or we were never given a token in the first place, fetch the ledger ID from
+            # the destination
             stub = lapipb.LedgerIdentityServiceStub(self._channel)
             response = await stub.GetLedgerIdentity(lapipb.GetLedgerIdentityRequest())
-            if isinstance(self._config.access, PropertyBasedAccessConfig):
-                self._logger.info("Connected to gRPC Ledger API, ledger ID: %s", response.ledger_id)
-                self._config.access.ledger_id = response.ledger_id
-            else:
-                raise ValueError("when using token-based access, the token must contain ledger ID")
+
+            self._logger.info("Connected to gRPC Ledger API, ledger ID: %s", response.ledger_id)
+            self._config.access.ledger_id = response.ledger_id
 
     async def close(self) -> None:
         """
@@ -674,6 +698,12 @@ class Connection(aio.Connection):
 
     # endregion
 
+    async def get_version(self) -> "Version":
+        stub = lapipb.VersionServiceStub(self.channel)
+        request = lapipb.GetLedgerApiVersionRequest(ledger_id=self._config.access.ledger_id)
+        response = await stub.GetLedgerApiVersion(request)
+        return Codec.decode_version(response)
+
     # region User Management calls
 
     async def get_user(self, user_id: "Optional[str]" = None) -> "User":
@@ -682,12 +712,27 @@ class Connection(aio.Connection):
         response = await stub.GetUser(request)
         return Codec.decode_user(response.user)
 
+    async def create_user(self, user: "User", rights: "Optional[Sequence[Right]]" = None) -> "User":
+        stub = lapiadminpb.UserManagementServiceStub(self.channel)
+        request = lapiadminpb.CreateUserRequest(user=Codec.encode_user(user))
+        if rights is not None:
+            request.rights.extend(Codec.encode_right(right) for right in rights)
+
+        response = await stub.CreateUser(request)
+        return Codec.decode_user(response.user)
+
     async def list_users(self) -> "Sequence[User]":
         stub = lapiadminpb.UserManagementServiceStub(self.channel)
         request = lapiadminpb.ListUsersRequest()
         response = await stub.ListUsers(request)
         return [Codec.decode_user(user) for user in response.users]
 
+    async def list_user_rights(self, user_id: "Optional[str]" = None) -> "Sequence[Right]":
+        stub = lapiadminpb.UserManagementServiceStub(self.channel)
+        request = lapiadminpb.ListUserRightsRequest(user_id=user_id)
+        response = await stub.ListUserRights(request)
+        return [Codec.decode_right(user) for user in response.rights]
+
     # endregion
 
     # region Party Management calls
